// -*- c++ -*-
#ifndef BTREE_H
#define BTREE_H

#include <strings.h>
#include <iostream.h>
#include "wfqueue.h"
#define BTREE_KEYS 4 
#define QBOTTOM  ((btree_op::ranger_queue*)-1)
#define QTOP     ((btree_op::ranger_queue*)0)

#define TBOTTOM  ((btree::op_queue::token)0)
#define TTOP  (-1)

#define RBOTTOM  ((btree::node)0)
#define RTOP  (-1)

class btree_op;
class btree_node;
class btree;


// ------------------------------------------
// The Btree datastructure 
// ------------------------------------------
class btree
{
public:

  typedef wfqueue<int> result_set;
  typedef wfqueue<btree_op> op_queue;
  typedef btree_node * node;	// Suitable for switching to handle
  typedef const btree_node *const_node;
  typedef btree_op   * op;

public:

  btree ();

  bool search (int v);
  bool range (int l, int h, result_set * r); 
  bool insert (int v);
  bool remove (int v);
  void print_tree(int hold_values[]);

  static btree* buildtree();
  
  bool next_op(btree_op&, bool all=false);

private:
  bool root_insert (btree_op *op, op_queue::token& rtok);
  bool root_range  (btree_op *op, op_queue::token& rtok);

  // Next work node
  node next_insert_node (btree_op&op);
  node next_range_node  (btree_op&op);
  
  // special root insert cases.
  node help_insert_root(node root, int v);
  node help_insert_child(node root, const_node child, int v);
  node help_insert_grand(node root, const_node child, const_node grand, int v);

  // 
  node internal_range (btree_op&, node current);
private:  
  btree_node *root_;
  op_queue ops_;

  friend ostream& operator << (ostream& o, btree&b) ;
};

// ------------------------------------------
// An operation on a btree node.
// ------------------------------------------
class btree_op 
{
public:
  enum { INSERT, SEARCH, RANGE };

  typedef wfqueue<btree_op> op_queue;
  struct ranger {
    op_queue *opq;
    op_queue::token token;

    ranger() {}
    ranger (wfqueue<btree_op> *q, wfqueue<btree_op>::token (tok)) 
      : opq(q), token(tok){}
  };
  typedef wfqueue<ranger> ranger_queue;


public:
  btree_op () {}
  btree_op (const btree_op&op) ;

  static btree_op insert (int k) { 
    return btree_op(INSERT, k, k); 
  }
  static btree_op range (int v,int u, btree::result_set *r) { 
    return btree_op(RANGE, v, u, r); 
  }
  static btree_op search(int v) { return btree_op(SEARCH,v); }
  static unsigned long sequencer;

  int seqnum;
  int operation;
  op_queue::token  move_token;	// Used to move 
  btree::node      tree_token;	// Used to swap a root pointer of a subtree
				// This must be the value used in CAS

  union {
    struct {
      int key, val;
    } i;
    struct {
      int low, high;
      ranger_queue *tokens;
      btree::result_set *results;
    } r;
    struct {
      int key;
    } s;
  } u;

  bool operator == (const btree_op&op) {return (seqnum == op.seqnum);}
  bool operator != (const btree_op&op) {return (seqnum != op.seqnum);}
  bool operator < (const btree_op&op) {return (seqnum < op.seqnum);}
  bool operator <= (const btree_op&op) {return (seqnum <= op.seqnum);}

  friend ostream& operator <<(ostream&o, const btree_op&op);
private:

  btree_op (int op, int key, int val) {
    init_op (op);
    u.i.key = (key); u.i.val = (val); 
  }

  btree_op (int op, int l, int h, btree::result_set * r) {
    init_op (op);
    u.r.low=l; u.r.high=h; u.r.results = r; u.r.tokens = QBOTTOM; 
  }
  
  btree_op (int op, int key) {
    init_op (op);
    u.s.key = (key); 
  }

  void init_op (int op) {
    seqnum = INC(sequencer);    
    operation = op;
    move_token = TBOTTOM;
    tree_token = RBOTTOM;
  }

};

// ------------------------------------------
// A node in the Btree
// Each node is a sequence of entries
// ------------------------------------------

struct node_entry {
  int        key;
  btree_node* kid;
};

class btree_node
{
  typedef btree_node *node;
  typedef const btree_node *const_node;

  typedef wfqueue<btree_op> op_queue;

public:
  btree_node ();
  ~btree_node ();

  void print_tree(int hold_values[]) {
   
    btree_op *peek_op;

 
    op_queue::token tok;

    
    if (isleaf()) {
      for (int j=0; j<keycount_; j++) 
	hold_values[key(j)] = key(j);
    }
    else {
      for (int i=0; i<keycount_; i++)
	kid(i)->print_tree(hold_values);
    }
  }

  node  clone () const { return new btree_node (*this); }
  node  clone2 (int child_index) const {
    node root = clone();
    root->kid (child_index) = root->kid(child_index)->clone();
    return root;
  }

  void insert (int v);
  void range (int v, int u, btree::result_set* r);
  bool search (int v);

  bool isleaf () const { return leaf_; } // Leaf have no children.
  bool isfull()  const {return keycount() >= BTREE_KEYS;}
  int  keycount ()   const { return keycount_; }
  void setleaf(bool v) { leaf_ = v; }
  void dump_level (ostream&o);

  int&  key(int index) { return v_[index].key; }
  int   key(int index) const { return v_[index].key; }
  node& kid(int index) { return v_[index].kid; }
  node  kid(int index) const { return v_[index].kid; }
  int&  val(int index) { return *((int*)&v_[index].kid); }
  node_entry& entry (int index) { return v_[index]; }

  // Insert help operations
  int  scan_keys (int key) const;
  int  scan_kids (const_node) const; 

  node find_kid (int key) const;

  void makesomespace (int index);
  void copyleftstuff (node tocopy) const ;
  void copyrightstuff (node tocopy) const;
  void insert_on_newleaf (const_node leaf, int v);

private:

  bool next_op(btree_op&, bool all=false);	// Remove next op from the queue
				// and perform the operation.

  node next_insert_node (btree_op&op);
  node help_insert_grand(const_node child, const_node grand, int v);

  bool internal_insert (btree_op*, op_queue::token& remtok);
  void internal_range (btree_op*, op_queue::token& remtok);
  //void internal_search (btree_op&, op_queue::token& remtok, bool mine);
  void insert_node(int v);
  // Access to internal stuff.
  //int  split_node (node& newkid);
  static bool split_child(node parent, 
				 const_node child,
				 node newleft,
				 node newright);

  static btree_node *split_root(btree_node *root, 
				btree_node *newleft,
				btree_node *newright);
 
private:
  friend class btree;
  
  op_queue *ops_;
  bool     leaf_;
  int      keycount_;
  struct node_entry   v_[BTREE_KEYS+1];

  friend ostream& operator<< (ostream& o, const btree_node&n);
};
inline
void btree::print_tree(int hold_values[]) {root_->print_tree(hold_values);};
#endif

